import { URL } from 'url';
export interface URLOptions {
    href?: string;
    protocol?: string;
    host?: string;
    hostname?: string;
    port?: string | number;
    pathname?: string;
    search?: string;
    searchParams?: unknown;
    path?: string;
}
declare const _default: (origin: string, options: URLOptions) => URL;
export default _default;
