import { Got, HandlerFunction, InstanceDefaults } from './types';
export declare const defaultHandler: HandlerFunction;
declare const create: (defaults: InstanceDefaults) => Got;
export default create;
export * from './types';
