export declare type DnsLookupIpVersion = 'auto' | 'ipv4' | 'ipv6';
declare type DnsIpFamily = 0 | 4 | 6;
export declare const isDnsLookupIpVersion: (value: any) => boolean;
export declare const dnsLookupIpVersionToFamily: (dnsLookupIpVersion: DnsLookupIpVersion) => DnsIpFamily;
export {};
