export default class WeakableMap<K, V> {
    weakMap: WeakMap<Record<string, unknown>, V>;
    map: Map<K, V>;
    constructor();
    set(key: K, value: V): void;
    get(key: K): V | undefined;
    has(key: K): boolean;
}
