import { CancelableRequest, BeforeErrorHook } from './types';
export default function createRejection(error: Error, ...beforeErrorGroups: Array<BeforeErrorHook[] | undefined>): CancelableRequest<never>;
