import { NormalizedOptions, Defaults } from './types';
declare const normalizeArguments: (options: NormalizedOptions, defaults?: Defaults | undefined) => NormalizedOptions;
export default normalizeArguments;
